// 7_4_EjemploDeConsola.cpp : Defines the entry point for the console application.
//

#include "stdafx.h"

#include <winsock2.h>
#pragma comment(lib, "WS2_32.LIB")

#include <new>
using namespace std;

#define PORT 23456

void PrintError(char* ProblemMessage, int ErrorCode)
{
	const char* ConnectionString = ": ";
	char* CompleteMessage;
	LPVOID ReasonMessage;
	DWORD ReasonMessageSize; 

	ReasonMessageSize = FormatMessage( 
		FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM
		| FORMAT_MESSAGE_IGNORE_INSERTS,
		NULL,
		ErrorCode,
		MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), 
		(LPTSTR) &ReasonMessage,
		0,
		NULL 
	);

	CompleteMessage = new(nothrow) char[lstrlen(ProblemMessage) 
		+ lstrlen(ConnectionString) + ReasonMessageSize + 1];
	if (!CompleteMessage)
		return;
	lstrcpy(CompleteMessage, ProblemMessage);
	lstrcat(CompleteMessage, ConnectionString);
	lstrcat(CompleteMessage, (char*)ReasonMessage);
	LocalFree(ReasonMessage);

	fprintf(stderr,CompleteMessage);

	delete[] CompleteMessage;
}

SOCKET CreateServer(unsigned short Port)
{
	SOCKADDR_IN	ServerAddr;
	SOCKET ListeningSocket;

	// Creacin del socket que escuchar a los clientes
	ListeningSocket = socket(
		AF_INET,		// Protocolo IPv4
		SOCK_STREAM,	// TCP/IP
		IPPROTO_TCP		// TCP
		);
	if (ListeningSocket == INVALID_SOCKET)
		return INVALID_SOCKET;

	ServerAddr.sin_family = AF_INET;
	// Puerto donde escuchar
	ServerAddr.sin_port = htons(Port);
	// Direccin por la que recibir
	ServerAddr.sin_addr.s_addr = htonl(INADDR_ANY);	

	// Asociacin de la direccin con el socket
	if (bind(
		ListeningSocket,			// Socket
		(SOCKADDR *)&ServerAddr,	// Direccin
		sizeof(ServerAddr)			// Tamao
	) == SOCKET_ERROR)
		return INVALID_SOCKET;

	if (listen(
		ListeningSocket, // Sockets
		SOMAXCONN	// Mximas conexiones a la espera
		) == SOCKET_ERROR)
		return INVALID_SOCKET;

	return ListeningSocket;
}

int RecvEx(SOCKET s, char* buf, int len, int flags)
{
	// Forma sencilla:
	// return recv(s, buf, len, flags);
	// Posible problema: no recibe el paquete de datos completo si este
	// se ha fragmentado

	// Versin mejorada
	// Solo para sockets bloqueantes y que esperen recibir un nmero de bytes
	// fijo
	int     ret, nLeft, idx;

	nLeft = len;
	idx = 0;

	while (nLeft > 0)
	{
		ret = recv(s, &buf[idx], nLeft, flags);
		if (ret == SOCKET_ERROR)
			return SOCKET_ERROR;
	    idx += ret;
		nLeft -= ret;
	}
	return len;
}

SOCKET CreateClient(unsigned short ServerPort, const char* ServerDir)
{
	SOCKADDR_IN	ServerAddr;
	SOCKET ServiceSocket;

	// Creacin del socket que se conectar al servidor
	ServiceSocket = socket(
		AF_INET,		// Protocolo IPv4
		SOCK_STREAM,	// TCP/IP
		IPPROTO_TCP		// TCP
		);
	if (ServiceSocket == INVALID_SOCKET)
		return INVALID_SOCKET;

	ServerAddr.sin_family = AF_INET;
	ServerAddr.sin_port = htons(ServerPort);
	ServerAddr.sin_addr.s_addr =  inet_addr(ServerDir);	

	// Si ServerDir es un nombre lgico hay que resolverlo
	if (ServerAddr.sin_addr.s_addr == INADDR_NONE)
	{
	    struct hostent *hp;
	    hp = (struct hostent *)gethostbyname(ServerDir);
	    if (hp == NULL)
			return INVALID_SOCKET;    
		else 
			memcpy((void *)&ServerAddr.sin_addr,
				(void *)hp->h_addr_list[0], 
				sizeof(ServerAddr.sin_addr));
	}

	if (connect(ServiceSocket, (struct sockaddr *)&ServerAddr, 
		sizeof(ServerAddr)) == SOCKET_ERROR)
		return INVALID_SOCKET;

	return ServiceSocket;
}

DWORD WINAPI ClientThread(LPVOID /*lpParam*/)
{
	SOCKET ServiceSocket;
	int DataNumber;

	printf("ClientThread: creating client...\n");
	ServiceSocket = CreateClient(PORT, "127.0.0.1");
	if (ServiceSocket == INVALID_SOCKET)
	{
		PrintError("ClientThread: Error in CreateClient", WSAGetLastError());
		return INVALID_SOCKET;
	}
	DataNumber = 3;

	printf("ClientThread: sending DataNumber = %d\n", DataNumber);
	if (send(ServiceSocket, (char*)&DataNumber, sizeof(DataNumber), 0) 
		== SOCKET_ERROR )
	{
		PrintError("ClientThread: Error in send", GetLastError());
		return INVALID_SOCKET;
	}

	if (RecvEx(ServiceSocket, (char*)&DataNumber, sizeof(DataNumber), 0)
		== SOCKET_ERROR )
	{
		PrintError("ClientThread: Error in RecvEx", GetLastError());
		return INVALID_SOCKET;
	}
	printf("ClientThread: received DataNumber = %d\n", DataNumber);

	closesocket(ServiceSocket);

	return 0;
}

DWORD WINAPI ServerThread(LPVOID /*lpParam*/)
{
	SOCKET ServiceSocket, ListeningSocket;
	SOCKADDR_IN	ClientAddr;
	int ClientAddrLen = sizeof(ClientAddr);
	int DataNumber;

	printf("ServerThread: creating server...\n");
	ListeningSocket = CreateServer(PORT);
	if (ListeningSocket == INVALID_SOCKET)
	{
		PrintError("ServerThread: Error in CreateServer", WSAGetLastError());
		return INVALID_SOCKET;
	}

	printf("ServerThread: waiting for client connection...\n");
	ServiceSocket = accept(
		ListeningSocket,	// Socket
		(SOCKADDR *)&ClientAddr,		// Direccin del cliente
		&ClientAddrLen					// Tamao
		);

	printf("ServerThread: client connected from %s:%d\n",
		inet_ntoa((struct in_addr)ClientAddr.sin_addr), ClientAddr.sin_port);

	if (RecvEx(ServiceSocket, (char*)&DataNumber, sizeof(DataNumber), 0)
		== SOCKET_ERROR )
	{
		PrintError("ServerThread: Error in RecvEx", GetLastError());
		return INVALID_SOCKET;
	}
	printf("ServerThread: received DataNumber = %d\n", DataNumber);

	DataNumber = DataNumber*DataNumber;

	printf("ServerThread: sending DataNumber = %d\n", DataNumber);
	if (send(ServiceSocket, (char*)&DataNumber, sizeof(DataNumber), 0) 
		== SOCKET_ERROR )
	{
		PrintError("ServerThread: Error in send", GetLastError());
		return INVALID_SOCKET;
	}
	closesocket(ServiceSocket);
	closesocket(ListeningSocket);

	return 0;
}

int _tmain(int /*argc*/, _TCHAR* /*argv[]*/)
{
	HANDLE hThreadClient, hThreadServer;
	WSADATA wsaData;

	// Inicializacin de WinSock 2.2
	if (WSAStartup(MAKEWORD(2,2), &wsaData) != 0)
	{
		PrintError("Error in WSAStartup", GetLastError());
		return -1;
	}

	hThreadServer = CreateThread(NULL, 0, ServerThread, NULL, 0, NULL);
	if (hThreadServer == NULL)
	{
		PrintError("Error in CreateThread", GetLastError());
		return -1;
	}

	// Dar tiempo a la creacin del servidor, esperar 1 segundo
	// Se podra mejorar utilizando eventos para no tener que esperar
	// un tiempo fijo
	Sleep(1000); 

	hThreadClient = CreateThread(NULL, 0, ClientThread, NULL, 0, NULL);
	if (hThreadClient == NULL)
	{
		PrintError("Error in CreateThread", GetLastError());
		return -1;
	}

	if (WaitForSingleObject(hThreadClient, INFINITE) == WAIT_FAILED)
	{
		PrintError("Error in WaitForSingleObject", GetLastError());
		return -1;
	}

	if (WaitForSingleObject(hThreadServer, INFINITE) == WAIT_FAILED)
	{
		PrintError("Error in WaitForSingleObject", GetLastError());
		return -1;
	}
	WSACleanup();

	return 0;
}

